/*
   This file is part of tiling.

   tiling is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   tiling is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tiling.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2010 Neil Birkbeck
*/
#ifndef NACB_HALF_EDGE_H
#define NACB_HALF_EDGE_H

#include <vector>
#include <nmath/vec2.h>

using namespace nacb;

class HalfEdgePolygon;

/**
   The HalfEdge is the data structure that makes up the
   edges of the HalfEdgePolygon.  These edges store and
   index of a point into a shared point list, and also
   store pointers to the next and previous HalfEdges 
   in the owner polygon, as well as a pointer to an
   opposite HalfEdge (which may be null for points on the boundary).
 */
class HalfEdge {
public:

  std::vector<Vec2d> &  points;
  HalfEdge * next, * prev;
  HalfEdge *  opp;
  HalfEdgePolygon  * poly;
  int point;

  Vec2d s1, s2;
  bool  smooth;

  /**
     Construct a new HalfEdge that will index the points, pts,
     using the index p for its point.

     \param pts  the shared list of points.
     \param  p   this half-edges index into the points.
  */
  HalfEdge(std::vector<Vec2d> & pts, int p) : points(pts), point(p) {
    next = opp = prev = 0;
    smooth = true;
  }

  ~HalfEdge() {

  }

  /**
     Set the next edge to be n, and fix up any backward links
     n must not be null.
     \param n  the next edge.
   */
  void setNext(HalfEdge * n) {
    next = n;
    n->prev = this;
  }

  /**
     Make the HalfEdge have a striaght edge.   
  */
  void makeStraight() {
    s1 = (points[point]*0.3333 + points[next->point]*0.6666);
    prev->s2 = (points[point]*0.666 + points[prev->point]*0.333);
    smooth = false;
  }

  /**
     Make the HalfEdge bezier points smooth (randomly)
     \param pstraight  the probability of being smooth
     \param minsmooth  the minimum smoothnes (smoothness will range from minsmooth to 1)
   */
  void makeSmooth(double pstraight, double minsmooth) {
    bool straight = false;
    if (straight) {
      makeStraight();
      return;
    }

    Vec2d d1 = points[next->point] - points[point];
    Vec2d d2 = points[point] - points[prev->point];

    double t = drand48()*(1.0-minsmooth) + minsmooth;
 
    if (drand48()<pstraight) {
      t=0;
      //smooth = false;
    }
    double d1len = d1.normalize();
    double d2len = d2.normalize();
    Vec2d dir = d1+d2;
    dir.normalize();

    double dmax = std::min(d1len, d2len)*0.5;
    if(d1.dot(d2)<0){
      dmax *= 1+d1.dot(d2);
    }
    s1 = points[point] + dir*dmax*t;
    prev->s2 = points[point] - dir*dmax*t;
  }

  /**
     Fix up the opposite HalfEdges Bezier control points to agree
     with this Edge.
  */
  void fixSmooth(){
    if(opp){
      double t = drand48();
      s1 = (s1*t + opp->s2*(1.0-t));
      s2 = (s2*t + opp->s1*(1.0-t));

      opp->s2 = s1;
      opp->s1 = s2;
      
      opp->smooth = smooth;
    }
  }

  /**
     Force the half edge to lie within the box specified by the
     lower and upper bounds.

     \param mn  the lower bound
     \param mx  the upper bound.
   */
  void box(const Vec2d & mn, const Vec2d & mx){
    s1 = (s1 - mn);
    s2 = (s2 - mn);
    s1.x /= (mx.x-mn.x);
    s1.y /= (mx.y-mn.y);
    
    s2.x /= (mx.x-mn.x);
    s2.y /= (mx.y-mn.y);
  }

};


#endif // HAVE_EDGE_H
