/*
   This file is part of tiling.

   tiling is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   tiling is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tiling.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2010 Neil Birkbeck
*/
#ifndef HALF_EDGE_POLYGON_H
#define HALF_EDGE_POLYGON_H

#include "HalfEdge.h"
#include <nmath/vec3.h>
#include <vector>
#include <map>
#include <stdio.h>

#ifdef HAVE_CAIRO
 #include <cairo.h>
 #include <cairo-svg.h>
#endif

using namespace std;
using namespace nacb;

/**
   The HalfEdgePolygon is a Polygon that consists of a 
   list of HalfEdges.  The npoints refers to the number
   of HalfEdges needed to be trafersed to get back
   to the start HalfEdge, hed.  The HalfEdges store
   an index into a set of points, which is shared among
   many polygons.  
*/
class HalfEdgePolygon {
public :
  HalfEdge * hed;
  int npoints;
  Vec3d color;
  std::vector<Vec2d> & points;

  /**
     Create a new polygon using the set of points, and a starting
     HalfEdge.
     
     \param pts  the common list of points
     \param h    the half edge that starts this polygon.
  */
  HalfEdgePolygon(std::vector<Vec2d> & pts, HalfEdge * h): points(pts) {
    color = Vec3d(drand48()*0.2+0.2, drand48()*0.5+0.5, drand48()*0.5+0.5);
    hed = h;
    hed->poly = this;

    HalfEdge * finger = hed->next;
    npoints = 1;

    while (finger != hed) {
      finger->poly = this;
      finger = finger->next;
      npoints++;

      if (npoints>1000) {
	fprintf(stderr, "Error closing the loop\n");
	exit(1);
      }
    }
  }

  /**
     Create a new Triangle using the set of points and a list of 
     indices into the points to be used for a triangle.
    
     \param pts  the common list of points
     \param   t  an array of indices into points.
  */
  HalfEdgePolygon(std::vector<Vec2d> & pts, int t[3]) : points(pts) {
    color = Vec3d(drand48()*0.2+0.2, drand48()*0.5+0.5, drand48()*0.5+0.5);

    hed = new HalfEdge(points, t[0]-1);
    hed->next = new HalfEdge(points, t[1]-1);
    hed->next->next = new HalfEdge(points, t[2]-1);

    HalfEdge * prev = hed->next->next;
    HalfEdge * finger = hed;
    for(int i=0; i<3; i++, finger=finger->next){
      finger->prev = prev;
      prev->next = finger;
      prev = finger;
      finger->poly = this;
    }
    npoints = 3;
  }
  
  /**
     \param  which  which HalfEdge to return.
     \return the n-th HalfEdge
   */
  HalfEdge * get(int which){
    which = which % npoints;
    HalfEdge * he = hed;

    for (int i=0; i<which; i++) {
      he = he->next;
    }
    return he;
  }

  /**
     Print an ASCII representation of the polygon to the file.
  */
  void print(FILE * f = stdout){
    if(!npoints)return;

    fprintf(f, "poly %d", npoints);
    for(int i=0; i<npoints; i++){
      HalfEdge * h = get(i);

      Vec2d p0 = h->prev->s2;
      Vec2d p1 = h->s1;
      
      if(!h->prev->smooth)// || (h->prev->opp && !h->prev->opp->smooth))
	p0 = points[h->prev->point]*0.25 + points[h->point]*0.75;
      if(!h->smooth)// || (h->opp && !h->opp->smooth))
	p1 = points[h->next->point]*0.25 + points[h->point]*0.75;
      
      fprintf(f, " %lf %lf %lf %lf %lf %lf",
	      p0.x, p0.y,
	      points[h->point].x, points[h->point].y,
	      p1.x, p1.y);
    }
    fprintf(f, "\n");
  }

#ifdef HAVE_CAIRO
  /**
     Draw the polygon to the cairo context using the given size.
     
     
     \param cairo  the context to draw to.
     \param sz     the size to draw this polygon.
  */
  void draw(cairo_t * cairo, double sz){
    if(!hed || !npoints)return;

    double thick = 1;

    //cairo_translate(cairo, 1, 1);
    //cairo_translate(cairo, )

    cairo_new_path(cairo);
    cairo_set_line_width(cairo, thick);
    cairo_set_source_rgb(cairo, color.x, color.y, color.z);

    HalfEdge * he = hed;
    
    cairo_move_to(cairo, points[he->point].x*sz, points[he->point].y*sz);
    for(int i=0; i<npoints; i++, he=he->next) {
      
      if(!he->smooth)
	cairo_line_to(cairo, points[he->next->point].x*sz, points[he->next->point].y*sz);
      else
	cairo_curve_to(cairo, 
		       he->s1.x*sz, he->s1.y*sz,
		       he->s2.x*sz, he->s2.y*sz,
		       points[he->next->point].x*sz, points[he->next->point].y*sz);
      //cairo_line_to(cairo, points[i1].pt.x, points[i1].pt.y);
    }
    cairo_close_path(cairo);
    //contours[i].draw(cairo);

    cairo_fill_preserve(cairo);

    cairo_set_source_rgb(cairo, 0.2,0.2,0.2);//line_color.x, line_color.y, line_color.z);
    cairo_stroke(cairo);

    he = hed;

    if(0)
    for(int i=0; i<npoints; i++, he = he->next){
      cairo_set_source_rgb(cairo, 0.5,0,0);

      cairo_new_path(cairo);
      Vec2d dir = points[he->next->point] - points[he->point];
      cairo_move_to(cairo, points[he->point].x*sz, points[he->point].y*sz);
      dir.normalize();
      dir*=10;
      Vec2d p = points[he->point]*sz + Vec2d(-dir.y, dir.x);
      cairo_line_to(cairo, p.x, p.y);
      cairo_stroke(cairo);

      if(!he->opp){
	cairo_new_path(cairo);
	cairo_set_source_rgb(cairo, 0.0,0.2,1.0);

	p = points[he->point] + points[he->next->point];
	
	p *= 0.5;
	p *= sz;

	cairo_move_to(cairo, p.x-dir.y, p.y+dir.x);
	cairo_line_to(cairo, p.x+dir.y, p.y-dir.x);
	cairo_stroke(cairo);
      }
    }
  }
#endif

  /**
     \return the number of HalfEdges that do not have an opposite
   */
  int missingOpp() {
    int n=0;
    for(int i=0; i<npoints; i++)
      n += (get(i)->opp == 0);
  }

  /**
     Make the polygon smooth

     \param pstraight  the probability of being straight
     \param minsmooth  the minimum smoothnes (smoothness will range from minsmooth to 1)
   */
  void makeSmooth(double pstraight, double minsmooth) {
    for(int i=0; i<npoints; i++)
      get(i)->makeSmooth(pstraight, minsmooth);
  }

  /**
     Make the smoothness (Bezier control points) of the opposite polygons agree
     with this polygons control points.
  */
  void fixSmooth() {
    for(int i=0; i<npoints; i++)
      get(i)->fixSmooth();
  }

  /**
     Intersect the line connecting points p1 and p2 with the
     implicit representation of a line.
     
     \param line  implicit representation of the line.
     \param p1    the first point.
     \param p2    the second point.
   */
  Vec2d isect(const Vec3d & line,
	      const Vec2d & p1,
	      const Vec2d & p2) const {
    /*
    Vec3d l2 = Vec3d(p1.x,p1.y,1.0).cross(Vec3d(p2.x,p2.y,1.0));
    Vec3d p = line.cross(l2);
    if(fabs(p.z)<=1e-5)
      return p1;
    return Vec2d(p.x/p.z, p.y/p.z);
    */
    double t1 = line.dot(Vec3d(p1.x, p1.y, 1.0));
    double t2 = line.dot(Vec3d(p2.x, p2.y, 1.0));
    if(fabs(t1) + fabs(t2) < 1e-7)
      return p1;
    return p1 + (p2-p1)*(fabs(t1)/(fabs(t1)+fabs(t2)));
  }

  /**
     \return 1 if the point is to the left of the line
  */
  bool side(const Vec3d & line, const Vec2d & pt){
    return line.dot(Vec3d(pt.x, pt.y, 1.0)) <= 0.0;
  }

  /**
     Split the polygon on the line.

     \param line     the line to split on (implicit representation)
     \param between  in/out tracking of new points that are created between
                     two existing points.  This map is indexed with a 
		     linear representation of the 2 point indexes (e.g.,
		     between[pt1*65536 + pt2] gives the index of a point added
		     between pt1 and pt2.  This is used to keep track of 
		     new points that are added, so when opposite triangles are split,
		     they will use the same new point.
    \return  a list of the new polygons
  */
  std::vector<HalfEdgePolygon *> split(const Vec3d & line,
				       map<int, int> & between){
    HalfEdge * he = hed;
    HalfEdge * switched = 0;
    int pt1 = 0;
    std::vector<HalfEdgePolygon *> pgons;
    int psize = 0;

    for(int i=0; i<npoints; i++){
      if(side(line, points[he->point]) !=
	 side(line, points[he->next->point])){
	int i2 = (he->point<<16) | he->next->point;

	if(between.count(i2) == 0){
	  Vec2d newpt = isect(line, 
			      points[he->point],
			      points[he->next->point]);
	  int i1 = (he->next->point<<16) | he->point;

	  between[i1] = points.size();
	  between[i2] = points.size();
	  points.push_back(newpt);
	}

	// Every second intersection a new polygon will be created that
	// links the previous (e.g., switched) half edge to the 
	// current intersection point.
	if (switched) {
	  HalfEdge * next = he->next;
	  HalfEdge * he1 = new HalfEdge(points, pt1);
	  HalfEdge * he2 = new HalfEdge(points, between[i2]);
	  HalfEdge * he2a = new HalfEdge(points, pt1);
	  HalfEdge * he1a = new HalfEdge(points, between[i2]);
	  HalfEdge * part = switched->next;

	  switched->opp = switched->next->opp = 0;
	  he->opp = he->next->opp = 0;

	  he1->poly = this;
	  he1a->poly = this;

	  switched->setNext(he1);
	  he1->setNext(he1a);
	  he1a->setNext(he->next);

	  he->setNext(he2);
	  he2->setNext(he2a);
	  he2a->setNext(part);
	  
	  HalfEdgePolygon * pgon = new HalfEdgePolygon(points, he);
	  
	  he1->makeStraight();
	  he2->makeStraight();
	  he1a->makeStraight();
	  he2a->makeStraight();
	  switched->makeStraight();
	  he->makeStraight();

	  switched = 0;
	  pgons.push_back(pgon);

	  he = next;
	  continue;
	}
	else {
	  switched = he;
	  pt1 = between[i2];
	}
      }
      he = he->next;
    }

    he = hed->next;
    npoints = 1;
    
    while (he != hed) {
      he = he->next;
      npoints++;
    }
    return pgons;
  }

  /**
    Remove and delete all half-edges from this polygon.
  */
  void clear() {
    HalfEdge * he = hed;
    for(int i=0; i<npoints; i++){
      HalfEdge * cur = he;
      he = he->next;
      he->opp = 0;
      delete cur;
    }
    hed = 0;
    npoints = 0;
  }

  /**
     Move the polygon off the edges.
   */
  void moveOffEdges(double eps = 1e-5) {
    int outside = 0;

    for(int i=0; i<npoints; i++){
      HalfEdge * he = get(i);
      Vec2d p1 = points[he->point];
      Vec2d p2 = points[he->next->point];
      if(he->opp)continue;
      if((p1.x<=eps || p1.x>=(1.0-eps) || p1.y<=eps || p1.y>=(1.0-eps)) &&
	 (p2.x<=eps || p2.x>=(1.0-eps) || p2.y<=eps || p2.y>=(1.0-eps))){

	if(drand48()<0.5){
	  double t = 0.7;
	  points[he->point] = points[he->point]*(1.0-t) + (he->s1 + he->prev->s2)*0.5*t;
	  points[he->next->point] = points[he->next->point]*(1.0-t) + (he->s2 + he->next->s1)*0.5*t;
	}
	double len = (p2-p1).len();
	Vec2d dir1 = (p1 - points[he->prev->point]);
	Vec2d dir2 = (p2 - points[he->next->next->point]);

	if(dir1.normalize()>=1e-7 && dir2.normalize()>=1e-7){
	  //Nothing wanted to work.  Just using direction of edge instead.
	  Vec2d dir(-(p2-p1).y, (p2-p1).x);
	  dir.normalize();
	  if(drand48()<0.5)
	    dir*=-1.0;
	  if(drand48()<0.8)
	    he->s1 = (p1*0.6 + p2*0.4)  + dir*len*drand48()*0.25;
	  if(drand48()<0.8)
	    he->s2 = (p2*0.6 + p1*0.4)  + dir*len*drand48()*0.25;
	  he->smooth = true;
	}
      }
    }
  }

  /**
     \param eps the tolerance in the computation
     \return the number of HalfEdges on the outside of the unit square using the
             the tolerance eps
   */
  int onOutside(double eps = 1e-5) {
    int outside = 0;
    for(int i=0; i<npoints; i++){
      Vec2d p1 = points[get(i)->point];
      outside += (p1.x<=eps || p1.x>=(1.0-eps) || p1.y<=eps || p1.y>=(1.0-eps));
    }
    return outside;
  }

  /**
     Clear the polygon when all of its points are outside (computed
     using the tolerance eps)
     \param eps  the tolerance to be used in the onOutside computation.
   */
  void removeOutside(double eps = 1e-5) {
    if(onOutside(eps) == npoints)
      clear();
  }

  /**
     After simplification, some of the HalfEdges will be degenerate lines.
     These lines are removed.  This should be called multiple times.
  */
  void removeDegen() {
    for(int i=0; i<npoints; i++){
      HalfEdge * he = get(i);
      if(he->opp == he->next){
	HalfEdge * next = he->next->next;
	HalfEdge * prev = he->prev;

	prev->next = next;
	next->prev = prev;

	if(hed == he)
	  hed = next;
	if(hed == he->opp)
	  hed = next;

	//delete he->opp;
	//delete he;
	
	npoints-=2;
	break;
      }
    }
  }

  /**
     Check the polygon for consistency in the pointers for the half-edges.
     Prints out warnings if consistency checks fail.

     \return true if consistency checks are satisfied.
  */
  bool check() {
    HalfEdge * he = hed;
    for (int i=0; i<npoints; i++, he = he->next) { 
      
      if (he->next->prev != he) {
	fprintf(stderr, "error in adjacencies\n");
	return false;
      }

      if (he->opp && (he->opp->next->point != he->point ||
		     he->next->point != he->opp->point)) {
	fprintf(stderr, "opposites are wrong!\n");
	return false;
      }

      if (he->poly != this) {
	fprintf(stderr, "polygon is incorrect\n");
	return false;
      }

      if (he->opp && he->opp->poly == he->poly) { 
	fprintf(stderr, "error, polygon opp is same as this %p (next:%p)\n",
		he->opp, he->next);
	return false;
      } 

      if (he->opp && he->opp->opp != he) {
	fprintf(stderr, "error in opps\n");
	return false;
      }
    }

    if (he != hed) { 
      fprintf(stderr, "error in number of points, didn't traverse\n");
      he = hed;
      for(int i=0; i<npoints; i++, he = he->next){
	fprintf(stderr, "%d ", he->point);
      }
      fprintf(stderr, " : %d\n", he->point);
      return false;
    }
    return true;
  }

  /**
     Make sure the polygon fits inside the box.
     \see HalfEdge::box
   */
  void box(Vec2d & mn, Vec2d & mx) {
    for (int i=0; i<npoints; i++)
      get(i)->box(mn, mx);
  }

  /**
     Compute the range of the polygons points using
     the running totals mn and mx.  mn and mx should be
     initialized to large and small values respectively 
     before calling this on a set of polygons.

     \param mn  the minimum coordinate (in/out)
     \param mx  the maximum coordinate (in/out)
  */
  void range(Vec2d & mn, Vec2d & mx) {
    for(int i=0; i<npoints; i++){
      int p = get(i)->point;
      mn = mn.min(points[p]);
      mx = mx.max(points[p]);
    }
  }
};


#endif // HAVE_EDGE_POLYGON_H
