/*
   This file is part of tiling.

   tiling is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   tiling is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tiling.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2010 Neil Birkbeck
*/
#include "PolygonTiling.h"


void PolygonTiling::computeTriangulation() {
  int ntri;

#ifdef USE_QHULL    
  char name[1024] = "/tmp/tiling-XXXXXX";
  mkstemp(name);
  FILE * f = fopen(name, "w");
    
  fprintf(f, "2 %d\n", npoints);
  for(int i=0; i<npoints; i++)
    fprintf(f, "%lf %lf\n", points[i].x, points[i].y);
  fclose(f);
    
    
  char cmd[1024];
  snprintf(cmd, 1024, "qhull i d < %s", name);
  FILE * tris = popen(cmd, "r");
  fscanf(tris, "%d", &ntri);
  for(int i=0; i<ntri; i++){
    int t[3];
    fscanf(tris, "%d %d %d", t, t+1, t+2);
    Vec2d e1 = points[t[1]-1] - points[t[0]-1];
    Vec2d e2 = points[t[2]-1] - points[t[0]-1];
    double zp = e1.x*e2.y - e1.y*e2.x;
    ///printf("zp %f\n", zp);
    t[0]++;
    t[1]++;
    t[2]++;
    polys.push_back(new HalfEdgePolygon(points, t));
  }
  fclose(tris);
  remove(name);
#else
    
  vector<Vec3<int> > ts = delaunay(points);
  ntri = ts.size();
  for(int i=0; i<ts.size(); i++){
    int t[3]={ts[i].x, ts[i].y, ts[i].z};
      
    Vec2d e1 = points[t[1]] - points[t[0]];
    Vec2d e2 = points[t[2]] - points[t[0]];
    double zp = e1.x*e2.y - e1.y*e2.x;
    //printf("zp %f\n", zp);
    if(zp<0){
      t[1] = ts[i].z;
      t[2] = ts[i].y;
    }
    t[0]++;
    t[1]++;
    t[2]++;
    polys.push_back(new HalfEdgePolygon(points, t));
  } 

#endif
  fixOpp();
}


/**
   Fix up any HalfEdges that don't have an opposite.
   This is inefficent right now O(n^2); could use a mapping.
*/
void PolygonTiling::fixOpp() {
  for (int i=0; i<polys.size(); i++) {
    for (int j=i; j<polys.size(); j++) {
      HalfEdge * he1 = polys[i]->hed;
      for (int k1=0; k1<polys[i]->npoints; k1++, he1 = he1->next) {
	if (he1->opp) continue;
	  
	HalfEdge * he2 = polys[j]->hed;
	for (int k2=0; k2<polys[j]->npoints; k2++, he2 = he2->next) {
	  if (he1->next->point == he2->point && 
	      he1->point == he2->next->point) {
	    he1->opp = he2;
	    he2->opp = he1;
	    break;
	  }
	}
      }
    }
  }

  int nop=0;
  for (int i=0; i<polys.size(); i++) {
    for (int j=0; j<polys[i]->npoints; j++)
      nop += !polys[i]->get(j)->opp;
  }
  fprintf(stderr, "without opp %d\n", nop);
}


void PolygonTiling::removeEdges(double rfrac) {
  int nedges = 0;
  for(int i=0; i<polys.size(); i++)
    nedges += polys[i]->npoints;
    
  // remove some edges
  for(int i=0; i<rfrac*nedges; i++){
    double tot = 0;
    vector<double> probs(polys.size());
      
    // Choose a probability to remove an edge from each polygon
    // that is proportional to 1/(npoints^2), which makes it more
    // likely to remove an edge that is connected to a triangle.
    for(int j=0; j<polys.size(); j++){
      double prob = 1.0/(polys[j]->npoints*polys[j]->npoints);
      probs[j] = prob;
      tot += prob;
    }
      
    double p = drand48()*tot;
    int pii = (int)(double(rand())/double(RAND_MAX)*double(polys.size()));
    tot = 0;
    for(int j=0; j<polys.size(); j++){
      tot += probs[j];
      if(p<tot){
	pii = j;
	break;
      }
    }
    int which = rand()%polys[pii]->npoints;
    ///printf("removing %d/%d\n", which, nedges);fflush(stdout);
      
    for(int pi=pii; pi<=pii; pi++){
      if(which>=polys[pi]->npoints)
	which-=polys[pi]->npoints;
      else {
	///printf("removing, pi is %d, which=%d\n", pi, which);
	  
	//remove polys->heds[which]
	HalfEdge * rem = polys[pi]->get(which);
	HalfEdge * opp = rem->opp;
	if(!opp)break;
	  
	//This is degenerate edge.
	if(rem->poly == opp->poly)break;
	  
	HalfEdgePolygon * goner = opp->poly;
	for(int k=0; k<goner->npoints; k++)
	  goner->get(k)->poly = rem->poly;
	  
	rem->prev->next = opp->next;
	opp->next->prev = rem->prev;
	  
	rem->next->prev = opp->prev;
	opp->prev->next = rem->next;
	  
	  
	rem->poly->npoints = rem->poly->npoints+goner->npoints-2;
	if(rem->poly->hed == rem)
	  rem->poly->hed = rem->next;
	  
	delete rem;
	delete opp;
	  
	goner->npoints = 0;
	goner->hed = 0;
	  
	for(int ii=0; ii<polys.size(); ii++){
	  for(int jj=0; jj<polys[ii]->npoints; jj++){
	    if(polys[ii]->get(jj) == rem ||
	       polys[ii]->get(jj) == opp) {
	      fprintf(stderr, "error, something still referencing hed %p,%d %p %p %d,%p\n",
		      polys[ii]->get(jj), jj, rem, opp, ii, polys[ii]);
	    }
	  }
	}
	  
	
	delete goner;
	

	fprintf(stderr, "removing\n");
	for(int ii=0; ii<polys.size(); ii++){
	  if(polys[ii] == goner){
	    polys[ii] = polys[polys.size()-1];
	    polys.pop_back();
	    break;
	  }
	}
	fprintf(stderr, "removed %d\n", polys.size());
	//std::remove(polys.begin(), polys.end(), goner);
	///fixOpp(polys);
	check(); // Debug, could cause this to be slower than necessary.
	break;
      }
    }
    nedges-=2;
  }
}  



void PolygonTiling::smooth(int its) {
  vector<Vec2d> pback(points.size(), Vec2d(0,0));
  vector<int> ns(points.size(), 0);
  
  for (int it=0; it<its; it++) {
    for (int i=0; i<polys.size(); i++) {
      for (int j=0; j<polys[i]->npoints; j++) {
	HalfEdge * he = polys[i]->get(j);
	pback[he->point] += points[he->next->point];
	pback[he->point] += points[he->prev->point];
	ns[he->point]+=2;
      }
    }

    for (int i=0; i<points.size(); i++) {
      pback[i] *= 1.0/ns[i];
      points[i] = points[i]*0.9 + pback[i]*0.1;
    }
  }
  boxWithAspect(aspect);
}

