/*
   This file is part of tiling.

   tiling is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   tiling is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tiling.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2010 Neil Birkbeck
*/
/**
   Author: Neil Birkbeck
   (C) 2008-2009.
 */
#ifndef POLYGON_TILING_H
#define POLYGON_TILING_H

#include <vector>
#include <map>
#include "HalfEdgePolygon.h"
#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_CAIRO
 #include <cairo.h>
 #include <cairo-svg.h>
#endif

#ifndef USE_QHULL
 #include "delaunay.h"
#endif

using namespace std;
using namespace nacb;


/**
   The PolygonTiling class contains a set of polygons that
   are a tiling of the plane. Thes are originally obtained by 
   delaunay triangulation.  The class has operations for simplifying
   the triangles into polygons, and also for clipping the polygons
   to fit in a particular rectangle.
  
   Most of the time the points are in the rectange [0, aspect] x [0, 1].
   With the exception of clipping which happens in a unit box.
*/
class PolygonTiling {

public:
  std::vector<Vec2d>       points;
  vector<HalfEdgePolygon *> polys;
  double aspect;

  PolygonTiling() {
    aspect = 1.0;
  }

  /**
     Add some random points to the tiling (call this before triangulating).
     
     \param npoints the number of random points to add to the unitialized tiling
   */
  void addRandomPoints(int npoints) {
    for(int i=0; i<npoints; i++){
      points.push_back(Vec2d(drand48(), drand48()));
    }
  }

  /**
     Add some border points (proportional to sqrt(npoints)) to the set 
     of points in the tiling.

     Should be called before triangulating.

     \param npoints  the number of added points is proportional to the sqrt of.
   */
  void addBorderPoints(int npoints) {
    points.push_back(Vec2d(drand48()*1e-4, drand48()*1e-4));
    points.push_back(Vec2d(1.0-drand48()*1e-4, 1.0-drand48()*1e-4));
    points.push_back(Vec2d(drand48()*1e-4, 1.0-drand48()*1e-4));
    points.push_back(Vec2d(1.0-drand48()*1e-4, drand48()*1e-4));
    
    for(int i=0; i<=ceil(sqrt(npoints)); i++){
      points.push_back(Vec2d(drand48(), 0.0 + drand48()*1e-4));
      points.push_back(Vec2d(drand48(), 1.0 - drand48()*1e-4));
      points.push_back(Vec2d(0.0 + drand48()*1e-4, drand48()));
      points.push_back(Vec2d(1.0 - drand48()*1e-4, drand48()));
    }
    // npoints = points.size();
  }

  /**
     Set the aspect of the triangles to be aspect.  This assumes
     the aspect was 1 before.
     
     \param aspect the aspect ratio.
  */
  void setAspect(double aspect) {
    this->aspect = aspect;

    for(int i=0; i<points.size(); i++)
      points[i].x *= aspect;
  }

  /**
     Fit the tiling into a unit box.  This finds the current bounds of all
     points and then squishes them to fit in a box whos origin is [0, 0]
     and whos extents goes to [aspect, 1]
     
     \param aspect the aspect of the box to fit into.
  */
  void boxWithAspect(double aspect = 1.0){
    Vec2d mins(aspect, 1);
    Vec2d maxs(0, 0);
    for(int i=0; i<polys.size(); i++)
      polys[i]->range(mins, maxs);
    
    //Make sure transform gets scaled to preserve aspect.
    maxs.x = (maxs.x - mins.x)/aspect + mins.x;
    
    
    for(int i=0; i<points.size(); i++){
      points[i] -= mins;
      points[i].x /= maxs.x-mins.x;
      points[i].y /= maxs.y-mins.y;
    }
    
    for(int i=0; i<polys.size(); i++)
      polys[i]->box(mins, maxs);
  }
  
  
  /**
     Compute the trianglution of the current set of points.
  */
  void computeTriangulation();
  
  /**
     Fix up any HalfEdges that don't have an opposite.
     This is inefficent right now O(n^2); could use a mapping.
  */
  void fixOpp();

  /**
     Remove a proportion of the edges.  This turns triangles into
     polygons, while maintaining that the polygons tile the plane.
     
     \param rfrac  the proportion of edges to remove.
  */
  void removeEdges(double rfrac);


  /**
     Smooth the polygons such that the points of the polygons move towards the
     average of their neighbors.  Ensures that the resulting points are within
     the bounds of the aspect-box.
  */
  void smooth(int its = 10);

  /**
     Make the polygons smooth
     \param pstraight  the probability of being smooth
     \param minsmooth  the minimum smoothnes (smoothness will range from minsmooth to 1)
   */
  void makeAndFixSmooth(double pstraight, double minsmooth) {
    for(int i=0; i<polys.size(); i++)
      polys[i]->makeSmooth(pstraight, minsmooth);
    
    for(int i=0; i<polys.size(); i++)
      polys[i]->fixSmooth();
  }
  
  /**
     Remove all degenerate polygons.  Actually just clears the polygons so
     they don't appear.
   */
  void removeDegenerate() {
    for(int i=0; i<polys.size(); i++){
      for(int j=0; j<4; j++)
	polys[i]->removeDegen();
    }
  }

  /**
     Smooth the boundary: first make some lines and clip on the border
     lines.  Then offset (and smooth) a random set of the edges on the border.
						
     \param bsmooth the fraction of boundary edges to smooth. 
  */
  void smoothBoundary(double bsmooth) {
    const double ln = 0.05;
    vector<Vec3d> lines;
    
    lines.push_back(Vec3d(0, 1,  -ln));
    lines.push_back(Vec3d(-1, 0, 1. - ln));
    lines.push_back(Vec3d(0, -1, 1. - ln));
    lines.push_back(Vec3d(1, 0, -ln));

    // Make unit square (so all this clipping will work), fix up aspect after.
    boxWithAspect(1.0);

    map<int, int> between;
   
    /**
	 Add some lines and clip and split on these.
    **/
    for (int k=0; k<4; k++) {
      std::vector<HalfEdgePolygon *> newpolys;
      
      for(int i=0; i<polys.size(); i++) {
	std::vector<HalfEdgePolygon *> sps = polys[i]->split(lines[k], between);
	for(int j=0; j<sps.size(); j++)
	  newpolys.push_back(sps[j]);	
      }
      
      for(int j=0; j<newpolys.size(); j++)
	polys.push_back(newpolys[j]);
    }
    
    // Remove the polygons that are completely off.
    for(int i=0; i<polys.size(); i++)
      polys[i]->removeOutside(ln + 5e-3);

    fixOpp();
    
    // For polygons that have at least one point on the border,
    // randomly move some of them off the edges.
    for (int i=0; i<polys.size(); i++) {
      if(polys[i]->onOutside(ln) && drand48()<bsmooth){
	polys[i]->moveOffEdges(ln);
      }
    }

    for(int i=0; i<polys.size(); i++)
      polys[i]->fixSmooth();

    // Fix aspect, as the above operations used a unit box.
    boxWithAspect(aspect);

  }

  
  void print() const {
    for(int i=0; i<polys.size(); i++){
      polys[i]->print();
      //fflush(stdout); 
      //Had some problems running the loca version on the mac; thought this was a problem, but I'm pretty sure it was due to stderr not going anywhere.
    }
  }
  
#ifdef HAVE_CAIRO
  void drawCairo(const std::string & fname, int sz = 256) {
    int ww = sz * aspect, wh = sz;

    cairo_surface_t * surf = 
      cairo_svg_surface_create(fname.c_str(), ww, wh);

    cairo_t * cairo = cairo_create(surf);
    cairo_identity_matrix(cairo);
    
    //cairo_scale(cairo, 256, 256);

    for(int i=0; i<polys.size(); i++)
      polys[i]->draw(cairo, wh);

    cairo_surface_finish(surf);
    cairo_surface_destroy(surf);

    cairo_destroy(cairo);
  }
#endif

  bool check() {
    for(int i=0; i<polys.size(); i++){
      if(!polys[i]->check())return false;
    }
    return true;
  }
};

#endif
