/*
   This file is part of tiling.

   tiling is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   tiling is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tiling.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2010 Neil Birkbeck
*/
/* 
   Simple interface for delaunay triangulation.
   
   Copyright (c) 2009, Neil Birkbeck      
*/
#include "delaunay.h"
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <list>
#include <nmath/vec3.h>
#include <nmath/vec2.h>
#include <stdarg.h>
#include <stdlib.h>

using namespace std;
using namespace nacb;

static void del_warn(const char * fmt, ...){
#ifdef DEL_WARN
  va_arg arg;
  va_begin(arg, fmt);
  fprintf(stderr,"warning:");
  vfprintf(stderr, fmt, arg);
  va_end(arg);
#endif
}

static void del_debug(const char * fmt, ...){
#ifdef DEL_DEBUG
  va_arg arg;
  va_begin(arg, fmt);
  vfprintf(stderr, fmt, arg);
  va_end(arg);
#endif
}

/** \class event_t containts both site events and circle events, determined
    by the boolean site_event.

    Delclaration is private.
*/
struct event_t{
  bool site_event;
  float y;
  float x;//only useful for resolving comparisons
  union {
    int point;
    circle_t * circle;
  };
  event_t(Vec2d & p, int ind){
    y = p.y;
    x = p.x;
    point = ind;
    site_event = true;
  }
  event_t(float y, circle_t * circ){
    this->y = y;
    x = 0;
    circle = circ;
    site_event = false;
  }
  bool operator<(const event_t & o) const {
    if(y==o.y && x==o.x && x)del_warn("possible duplicate point\n");
    if(y==o.y)return x<o.x;
    return y<o.y;
  }
  bool operator<=(const event_t & o){
    if(y==o.y && x==o.x && x)del_warn("possible duplicate point\n");
    return y<=o.y;
  }
  bool false_alarm(){
    return !site_event && circle->false_alarm;
  }
};

inline double sqr(double d){return d*d;}
inline double sign(double d){return (d>0)?1:((d<0)?-1:0);}

double eval(double liney, const Vec2d & pi, double x){
  return  (sqr(x-pi.x) + sqr(pi.y) - sqr(liney))/(2.*(pi.y-liney));
}

/** Solve the quadratic equation and return appropriate solution given
    order of p1 and p2.
    \param p1 The site location of the first site arc
    \param p2 The site location of the second sites arc
    \return The x coordinate of solution.
*/
double solve(double yl,
	     const Vec2d & p1, 
	     const Vec2d & p2){
  double D1 = 2.0*(p1.y-yl);
  double D2 = 2.0*(p2.y-yl);

  double a = D2-D1;
  double b = 2.0*D1*p2.x - 2.0*D2*p1.x;
  double c =  D2*(p1.x*p1.x + p1.y*p1.y - yl*yl) 
            - D1*(p2.x*p2.x + p2.y*p2.y - yl*yl);
  double det = std::max(0.0, b*b-4.0*a*c);
  
  double t = -0.5*(b+sign(b)*sqrt(det));
  double s1 = t/a;
  double s2 = c/t;
  
  //double s1 = (-b + sqrt(det))/(2.0*a);
  //double s2 = (-b - sqrt(det))/(2.0*a);

  double res;
  if((res = fabs(a*s1*s1+b*s1+c))>1e-7){
    del_warn("quadratic solution has large residual: %g, s1=%f, s2=%f\n", res, s1, s2);
  }
  if((res = fabs(a*s2*s2+b*s2+c))>1e-7){
    del_warn("quadratic solution has large residual: %g, s1=%f, s2=%f\n", res, s1, s2);
  }

  if(p1.y > p2.y)return std::min(s1, s2);
  return std::max(s1, s2);
}

/** Intersect the two lines defined by pa+t1*da and pb+t2*db.
    \param pa point on first line.
    \param da direction of first line.
    \param pb point on second line.
    \param db direction of second line.
*/
Vec2d intersect(const Vec2d & pa,
	       const Vec2d & da,
	       const Vec2d & pb,
	       const Vec2d & db,
	       double & t1, double & t2){
  Vec2d diff = pb-pa;
  double det = (-da.x*db.y) - (-da.y*db.x);
  Vec2d r1(-db.y/det, db.x/det);
  Vec2d r2(-da.y/det, da.x/det);

  t1 = r1.dot(diff);
  t2 = r2.dot(diff);

  Vec2d pp1 = da*t1 + pa;
  Vec2d pp2 = db*t2 + pb;

  if(fabs(det)<1e-20){
    double dot = 1.0 - fabs(da.dot(db)/da.len()/db.len());
    del_warn("det in intersection is low: %lg, 1-fabs(dot): %lg\n", det, dot);
  }

  double res;
  if((res = (pp1-pp2).len())>1e-7){
    del_warn("intersection has large residual: %g\n", res);
  }

  return (pp1+pp2)*0.5;
}

/** Return the center of the circumcircle for the three points.
    \param p1 Point 1.
    \param p1 Point 2.
    \param p1 Point 3.
    \return The center of the circumcircle of p1,p2,p3.
*/
Vec2d center(const Vec2d & p1,
	     const Vec2d & p2,
	     const Vec2d & p3){
  Vec2d d1 = (p2-p1),     d2 = (p3-p2);
  Vec2d da(-d1.y, d1.x),  db(-d2.y, d2.x);
  Vec2d pa = (p1+p2)*0.5, pb = (p2+p3)*0.5;
  double t1, t2;
  return intersect(pa, da, pb, db, t1, t2);
}

/** Boolean predicate to test if beachlines of neighbors p1, p2, p3 will converge.
    \param p1 First beach arc.
    \param p2 Second beach arc.
    \param p3 Third beach arc.
    \param lower Output parameter for lower coordinate of circumcircle.
    \return True if breakpoints between lines will converge in future. 
*/
bool converge(double liney,
	      const Vec2d & p1,
	      const Vec2d & p2,
	      const Vec2d & p3,
	      double * lower = 0){

  Vec2d pp = center(p1, p2, p3);
  double r = (p1-pp).len();
  double low = pp.y - r;
  
  if(lower)*lower = low;

  double sol1 = solve(low, p1, p2);
  double sol2 = solve(low, p2, p3);
  
  if(isnan(sol1) || isinf(sol1) || isnan(sol2) || isinf(sol2))
    fprintf(stderr, "error: solution is nan %lf %lf\n", sol1, sol2);
  
  return fabs(sol1-sol2)<=1e-7;
}


vector<Vec3<int> > delaunay(vector<Vec2d> & pts,
			    interactions_func_t interactions){
  vector<event_t>       p;
  vector<segment_t>  segs;
  vector<Vec3<int> > tris;
  vector<circle_t *> delete_circles; //Delete these after (they may be used in the interactions function after deleted if deleted on the fly)
  double            sweep = 1;
  double             next = 1;

  for(int i=0; i<pts.size(); i++)
    p.push_back(event_t(pts[i], i));

  std::make_heap(p.begin(), p.end());

  if(p.size()){
    next = sweep = p.front().y;
  }

  while(p.size()){
    if(interactions){
      if(!interactions(0, delaunay_state_t(&segs, &tris, sweep, next)))
	break;
    }

    pop_heap(p.begin(), p.end());
    event_t ev = p.back();
    p.pop_back();

    sweep = ev.y;

    if(ev.site_event){
      if(interactions){
	if(!interactions(1, delaunay_state_t(&segs, &tris, sweep, next, ev.point)))
	  break;
      }

      del_debug("popped site event: %f %f\n", pts[ev.point].x, pts[ev.point].y);
	  
      if(segs.size()==0)
	segs.push_back(segment_t(ev.point));
      else {
	int fnd = segs.size()-1;

	//consistency check the beach line.
#ifdef DEL_CONSISTENCY
	vector<double> breakp;
	double last = -10e10;
	for(int i=0; i<int(segs.size()-1); i++){
	  double cur = solve(sweep, pts[segs[i].point], pts[segs[i+1].point]);
	  if(cur<=last){
	    del_warn("error in beach line: %lf >= %lf seg:%d, point:%d  circle:%p\n", last, cur, i, segs[i].point, segs[i].circle);
	    if(!segs[i].circle){
	      Vec2d p1 = pts[segs[i-1].point];
	      Vec2d p2 = pts[segs[i].point];
	      Vec2d p3 = pts[segs[i+1].point];
	      double low;
	      del_warn("warning: converging? %d %d %d %d\n", 
		       converge(sweep+0.1, p1, p2, p3, &low),
		       converge(sweep+0.01, p1, p2, p3, &low),
		       converge(sweep+0.001, p1, p2, p3, &low),
		       converge(sweep+0.0001, p1, p2, p3, &low));
		      
	    }
	    for(int j=std::max(1, i-2); j<std::min(i+2, int(segs.size()-1)); j++){
	      if(segs[j].circle && 
		 (segs[j].circle->p1 == segs[i].point ||
		  segs[j].circle->p2 == segs[i].point ||
		  segs[j].circle->p3 == segs[i].point)){
		fprintf(stderr, "it is in a circle at index %d (i=%d)\n", j, i);
	      }
	    }
	  }
	  double val = 10e10;
	  for(int j=0; j<int(segs.size()-1); j++){
	    double len = pts[segs[j+1].point].x - pts[segs[j].point].x;
	    if(len>1e-10 && j!=i && j!=(i+1)){
	      double v = eval(sweep, pts[segs[j].point], cur);
	      if(v>=sweep)
		val = std::min(val, v);
	    }
	  }
	  if(val +1e-4 < eval(sweep, pts[segs[i].point], cur)){
	    del_warn("val=%f<%f=cur\n", val, eval(sweep, pts[segs[i].point], cur));
	  }
	  last = cur;
	}
#endif //done consistency check.
	

	del_debug("break:");
	for(int i=0; i<int(segs.size()-1); i++){
	  del_debug("(%d %d  %f) ", segs[i].point,
		    segs[i+1].point,
		    solve(sweep, pts[segs[i].point], pts[segs[i+1].point]));
	}
	del_debug("\n");

	for(int i=0; i<int(segs.size()-1); i++){
	  double s = solve(sweep, pts[segs[i].point], pts[segs[i+1].point]);
	  if(pts[ev.point].x < s){
	    fnd = i;
	    break;
	  }
	}
	del_debug("inserting at %d\n", fnd);
	//add a new segment at fnd.
	
	del_debug("edge %d-%d\n", segs[fnd].point, ev.point);
	if(segs[fnd].circle){
	  del_debug("destroying circle event\n");

	  segs[fnd].circle->false_alarm = true;
	  segs[fnd].circle = 0;
	}

	segs.insert(segs.begin()+fnd, segs[fnd]);
	segs.insert(segs.begin()+fnd+1, segment_t(ev.point));
	fnd++;

	del_debug("break(after ): ");
	for(int i=0; i<int(segs.size()-1); i++){
	  del_debug("(%d: %d  %f) ", segs[i].point, 
		    segs[i+1].point,
		    solve(sweep, pts[segs[i].point], pts[segs[i+1].point]));
	}
	del_debug("\n");


	for(int i=0; i<segs.size(); i++){
	  del_debug("%d ", segs[i].point);
	}
	del_debug("  size:%d\n", segs.size());

	for(int j = std::max(fnd-1, -(fnd-2)); j<std::min((int)segs.size()-1, fnd+2); j++){
	  int p1 = segs[j-1].point;
	  int p2 = segs[j].point;
	  int p3 = segs[j+1].point;
	  if(p1 != p2 && p2 != p3 && p1 != p3){
	    double lower;
	    bool conv = converge(sweep, pts[p1], pts[p2], pts[p3], &lower);
	    if(conv){
	      if(segs[j].circle){
		del_debug("seg[].point=%d already has a circle: %d %d %d  adding: %d,%d,%d\n", segs[j].point, segs[j].circle->p1, segs[j].circle->p2, segs[j].circle->p3, p1, p2, p3);
		segs[j].circle->false_alarm = true;
	      }
	      segs[j].circle = new circle_t(p1,p2,p3);
	      p.push_back(event_t(lower, segs[j].circle));
	      push_heap(p.begin(), p.end());
	    }
	    del_debug("testing %d:  %d-%d-%d  %d\n", j, segs[j-1].point, segs[j].point, segs[j+1].point, conv);
	  }
	}
      }
    }
    else {
      if(ev.false_alarm()){
	del_debug("false alarm: %d-%d-%d\n", ev.circle->p1, ev.circle->p2, ev.circle->p3);
	delete_circles.push_back(ev.circle);
	continue;
      }
      del_debug("popped circle event: %d\n", ev.circle->p2);

      for(int i=0; i<segs.size(); i++){
	if(segs[i].circle == ev.circle){
	  if(i>0 && segs[i-1].circle){
	    del_debug("neighbor %d %d %d of segment %d\n", 
			segs[i-1].circle->p1,
			segs[i-1].circle->p2,
			segs[i-1].circle->p3, segs[i].point);

	    segs[i-1].circle->false_alarm = true;
	    segs[i-1].circle = 0;
	  }
	  if(i+1<segs.size() && segs[i+1].circle){
	    del_debug("neighbor %d %d %d of segment %d\n", 
		      segs[i+1].circle->p1,
		      segs[i+1].circle->p2,
		      segs[i+1].circle->p3, segs[i].point);
		   
	    segs[i+1].circle->false_alarm = true;
	    segs[i+1].circle = 0;
	  }
	  del_debug("circle %d %d %d\n", segs[i].circle->p1,
		    segs[i].circle->p2, segs[i].circle->p3);
	  tris.push_back(Vec3<int>(segs[i].circle->p1, segs[i].circle->p2, segs[i].circle->p3));
	  
	  del_debug("erasing segment index:%d, pt:%d mid:%d\n",i, segs[i].point, ev.circle->p2);
	  int erased = i;
	  segs.erase(segs.begin()+i);

	  del_debug("remaining:");
	  for(int j=0; j<segs.size(); j++)del_debug("%d ", segs[j].point);
	  del_debug("\n");
	  del_debug("found circle segment\n");

	  for(int j=i-1; j<=i; j++){
	    if(j>0 && j+1<segs.size()){
	      int p1 = segs[j-1].point;
	      int p2 = segs[j  ].point;
	      int p3 = segs[j+1].point;

	      if(p1 != p2 && p2 != p3 && p1 != p3){
		double lower;
		bool conv = converge(sweep, pts[p1], pts[p2], pts[p3], &lower);
		if(conv){
		  if(segs[j].circle)del_debug("already has a circle (second)\n");
		  del_debug("new potential circle %d %d %d\n", p1, p2, p3);
		  segs[j].circle = new circle_t(p1,p2,p3);
		  p.push_back(event_t(lower, segs[j].circle));
		  push_heap(p.begin(), p.end());
		}
		del_debug("testing %d:  %d-%d-%d  %d\n", j, segs[j-1].point, segs[j].point, segs[j+1].point, conv);
	      }
	    }
	  }
	  break;
	}
      }
      delete_circles.push_back(ev.circle);
    }

    if(p.size())
      next = p.front().y;
    else
      next = -10e10;
      
    if(interactions){
      if(!interactions(3, delaunay_state_t(&segs, &tris, sweep, next)))
	break;
    }
  }

  for(int i=0; i<delete_circles.size(); i++)
    delete delete_circles[i];

  return tris;
}

