/*
   This file is part of tiling.

   tiling is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   tiling is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with tiling.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2010 Neil Birkbeck
*/
/* Copyright (c) 2008, Neil Birkbeck      
   
   This is the main program to compute a tiling of the plane.
   Uses a delaunay triangulation, and then simplifies a bunch
   of the triangles to make polygons.
*/

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>

#include <nmath/vec2.h>
#include <nmath/vec3.h>

#include <vector>
#include <map>
#include <string>

#include "HalfEdge.h"
#include "HalfEdgePolygon.h"
#include "PolygonTiling.h"

int main(int ac, char * av[]){  
  struct option opts[]={{"npoints", 1, 0, 'n'},
			{"minsmooth", 1, 0, 'm'},
			{"straight", 1, 0, 's'},
			{"frac", 1, 0, 'f'},
			{"bsmooth", 1, 0, 'b'},
			{"aspect", 1, 0, 'a'},
			{0, 0, 0, 0}};
  std::string fname = "tiled.svg";
  double pstraight = 0.5;
  double minsmooth = 0.5;
  double rfrac = 2.0/3.0;
  int npoints = 50, ntri = 0;  
  int ind, longind;
  double bsmooth = 0.0;
  double aspect = 1.0;

  // Process the command l
  while ((ind = getopt_long(ac, av, "n:m:s:f:b:a:", opts, &longind)) >= 0) {
    switch(ind){
    case 'n':
      npoints = atoi(optarg);
      break;

    case 'm':
      minsmooth = atof(optarg);
      break;

    case 's':
      pstraight = atof(optarg);
      break;

    case 'f':
      rfrac = atof(optarg);
      rfrac = std::max(0.01, std::min(rfrac, 0.99));
      break;

    case 'a':
      aspect = atof(optarg);
      break;

    case 'b':
      bsmooth = std::max(0.0, std::min(1.0, atof(optarg)));
      break;
    }
  }

  if (optind < ac) {
    fname = av[optind];
    if(fname.find("svg") == string::npos)
      fname += ".svg";       
  }

  srand(111);
  //srand(getpid());
  
  PolygonTiling polys;
  polys.addRandomPoints(npoints);
    
  if (bsmooth<1.0)
    polys.addBorderPoints(npoints);

  polys.setAspect(aspect);
  polys.computeTriangulation();

  polys.removeEdges(rfrac);
  polys.smooth();

  polys.makeAndFixSmooth(pstraight, minsmooth);
  polys.removeDegenerate();

  if (bsmooth<1.0) {
    polys.smoothBoundary(bsmooth);
  }

#ifdef HAVE_CAIRO
  // printf("file %s\n", fname.c_str());
  polys.drawCairo(fname.c_str());
  
#endif
  //add some edges

  polys.print();
  return 0;
}
